@@ -1,3 +1,15 @@
+'''
+get_groups_list
+    resolve_fname
+        load
+            load_file
+                as_valid_path
+            get_matched_categories
+                close_matches
+                    as_valid_path
+getallcategories
+
+'''
 import os
 import glob
 import fnmatch
@@ -21,8 +33,27 @@ ALIASES = {
     'adj': 'adjectives',
     'nn': 'nouns',
     'vb': 'verbs',
+    'u': 'uuid',
+    'uu': 'uuid',
 }
 
+import uuid
+def uuid_(n=None):
+    return str(uuid.uuid4())[:n and int(n)]
+WORD_FUNCS = {
+    'uuid': uuid_,
+}
+
+
+def get_groups_list(fnames):
+    '''Get word lists from multiple word groups.'''
+    return [x for f in as_multiple(fnames) for x in resolve_fname(f)]
+
+
+def resolve_fname(fname):
+    '''Detect if fname is a path or is a literal word. Replace any path
+    shortcuts.'''
+    return load(fname) if '/' in fname else [fname]
 
 
 def load(name):
@@ -38,46 +69,33 @@ def load(name):
 @functools.lru_cache(128)
 def load_file(name):
     '''Load a wordlist. Does not do any name conversion.'''
-    path = os.path.join(WORD_PATH, safepath(name + '.txt'))
-    if not os.path.isfile(path):
-        raise OSError("Wordlist '{}' does not exist.".format(name))
+    parts = name.split('/')
+    if parts[0] in WORD_FUNCS:
+        return [functools.partial(WORD_FUNCS[parts[0]], *(p for p in parts[1:] if p != '*'))]
 
-    with open(path, 'r') as f:
+    with open(as_valid_path(name, required=True), 'r') as f:
         return [
             l for l in (l.strip() for l in f.readlines())
             if l and l[0] not in ';#']
 
 
-def resolve_fname(fname):
-    '''Detect if fname is a path or is a literal word. Replace any path
-    shortcuts.'''
-    return load(fname) if '/' in fname else [fname]
-
-
-def get_groups_list(fnames):
-    '''Get word lists from multiple word groups.'''
-    return [x for f in as_multiple(fnames) for x in resolve_fname(f)]
-
-
-def getallcategories(d=''):
-    '''Get all categories (subdirectories) from a word class (adjectives,
-    nouns, etc.)'''
-    d = os.path.join(WORD_PATH, d)
-    return [
-        os.path.relpath(os.path.splitext(f)[0], d)
-        for f in glob.glob(os.path.join(d, '**/*.txt'), recursive=True)]
-
-
 @functools.lru_cache(128)
 def get_matched_categories(name, *a, **kw):
     '''Resolve a fuzzy-matched category. Throw an error if it doesn't
     match anything.'''
     name = doalias(name)
     # glob matching
+    if name.endswith('/'):
+        name += '*'
     matches = [f for f in ALL_CATEGORIES if fnmatch.fnmatch(f, name)]
-    matches = [name]
     if matches:
         return matches
+
+    parts = name.split('/', 1)
+    matches = ['/'.join((f, parts[1])) for f in WORD_FUNCS if fnmatch.fnmatch(f, parts[0])]
+    if matches:
+        return matches
+
     # no matches. throw nice error
     matches = close_matches(name, *a, **kw)
     raise ValueError("No matching wordlist '{}'. {}".format(
@@ -90,6 +108,7 @@ def close_matches(name, cutoff=0.65):
     # they entered a underspecified category
     name = doalias(name)
     matches = [cat for cat in ALL_CATEGORIES if name == cat.split('/', 1)[1]]
+    all_sub_categories = [c.split('/', 1)[-1] for c in ALL_CATEGORIES]
 
     if '/' in name:
         part0, part1 = name.split('/', 1)
@@ -98,18 +117,19 @@ def close_matches(name, cutoff=0.65):
         # they spelled the first part correctly
         if part0 in WORD_CLASSES:
             _ms = _get_matches(part1, AVAILABLE[part0], cutoff=cutoff)
-            matches += ['{}/{}'.format(part0, m) for m in _ms]
+            matches += [f for f in ('{}/{}'.format(part0, m) for m in _ms) if as_valid_path(f)]
         # they entered a misspelled category
-        elif part1 in ALL_CATEGORIES:
-            _ms = _get_matches(part1, ALL_CATEGORIES, cutoff=cutoff)
-            matches += ['{}/{}'.format(part0, m) for m in _ms]
+        elif part1 in all_sub_categories:
+            _ms = _get_matches(part0, [k for k in AVAILABLE if part1 in AVAILABLE[k]], cutoff=cutoff)
+            matches += [f for f in ('{}/{}'.format(m, part1) for m in _ms) if as_valid_path(f)]
         # they entered a misspelled category and misspelled group
         else:
-            matches += difflib.get_close_matches(name, ALL_CATEGORIES, cutoff=cutoff)
+            matches += _get_matches(name, ALL_CATEGORIES, cutoff=cutoff)
+        _ms = _get_matches(part0, list(WORD_FUNCS), cutoff=cutoff)
+        matches += ['{}/{}'.format(m, part1) for m in _ms]
     else:
         # get sub matches
-        _avail = [cat.split('/', 1)[1] for cat in ALL_CATEGORIES]
-        _ms = _get_matches(name, _avail, cutoff=cutoff)
+        _ms = _get_matches(name, all_sub_categories, cutoff=cutoff)
         matches += [
             '{}/{}'.format(pos, cat)
             for cat in _ms for pos in find_parts_of_speech(cat)]
@@ -171,6 +191,15 @@ def safepath(f):
     return os.path.abspath('/' + f).lstrip('/')
 
 
+def as_valid_path(name, required=False):
+    path = os.path.join(WORD_PATH, safepath(name + '.txt'))
+    if not os.path.isfile(path):
+        if required:
+            raise OSError("Wordlist '{}' does not exist.".format(name))
+        return
+    return path
+
+
 def prefix(pre, xs):
     '''Prefix all items with a path prefix.'''
     return [os.path.join(pre, x.lstrip('/')) for x in as_multiple(xs)]
@@ -179,13 +208,33 @@ def prefix(pre, xs):
 def choose(items, n=None):
     '''Choose one item from a list.'''
     items = as_multiple(items)
-    return random.choice(items) if n is None else random.choices(items, k=n)
+    x = random.choice(items) if n is None else random.choices(items, k=n)
+
+    if isinstance(x, list):
+        x = [xi() if callable(xi) else xi for xi in x]
+    elif callable(x):
+        x = x()
+    return x
+
+def sample_unique(func, n, *a, n_fails=50, unique=True, **kw):
+    if not unique:
+        return [func(*a, **kw) for _ in range(n)]
+    words = set()
+    for i in range(n):
+        for j in range(n_fails):
+            words.add(func(*a, **kw))
+            if len(words) > i:
+                break
+        else:
+            break
+    return words
 
 
 def as_multiple(x):
     '''Ensure a list or tuple.'''
     x = x if isinstance(x, (list, tuple)) else [x]
-    return [si for s in (str(s) for s in x) for si in s.split(',')]
+    return [si for s in x for si in ([s] if callable(s) else str(s).split(','))]
+
 
 
 def find_parts_of_speech(name):
@@ -193,6 +242,15 @@ def find_parts_of_speech(name):
     return {k for k, v in AVAILABLE.items() if name in v}
 
 
+def getallcategories(d=''):
+    '''Get all categories (subdirectories) from a word class (adjectives,
+    nouns, etc.)'''
+    d = os.path.join(WORD_PATH, d)
+    return [
+        os.path.relpath(os.path.splitext(f)[0], d)
+        for f in glob.glob(os.path.join(d, '**/*.txt'), recursive=True)]
+
+
 # get all available word classes and categories.
 AVAILABLE = {k: getallcategories(k) for k in WORD_CLASSES}
 ALL_CATEGORIES = getallcategories()  # FIXME: redundant
